/*
 * In the script the swarm follows continuously a list of checkpoint while
 * keeping the shape of a "V". This is a demo for the simulator.
*/


/*jslint nomen: true, indent: 2, maxerr: 3, maxlen: 80 */
/*global console, me*/

(function (console, me) {
  "use strict";

  var FLIGH_ALTITUDE = 100,
    MINIMAL_DISTANCE = 20,
    EPSILON = 20,
    CHECKPOINT_LIST = [
      {
        altitude: 585.1806861589965,
        latitude: 45.64492790560583,
        longitude: 14.25334942966329
      },
      {
        altitude: 589.8802607573035,
        latitude: 45.64316335436476,
        longitude: 14.26332880184475
      },
      {
        altitude: 608.6648153348965,
        latitude: 45.64911917196595,
        longitude: 14.26214792790128
      },
      {
        altitude: 606.1448368129072,
        latitude: 45.64122685351364,
        longitude: 14.26590493128597
      },
      {
        altitude: 630.0829598206344,
        latitude: 45.64543355564817,
        longitude: 14.27242391207985
      },
      {
        altitude: 616.1839898415284,
        latitude: 45.6372792927328,
        longitude: 14.27533492411138
      },
      {
        altitude: 598.0603137354178,
        latitude: 45.64061299543953,
        longitude: 14.26161958465814
      },
      {
        altitude: 607.1243119862851,
        latitude: 45.64032340702919,
        longitude: 14.2682896662383
      }
    ],
    R = 6371e3;

  function exitOnFail(ret, msg) {
    if (ret) {
      console.log(msg);
      me.exit(1);
    }
  }

  function deployParachute(drone) {
    console.log("[DEMO] Deploying parachute...");
    me.sendMsg("landing");
    exitOnFail(drone.triggerParachute(), "Failed to deploy parachute");
    me.exit(0);
  }

  function distance(lat1, lon1, lat2, lon2) {
    var la1 = lat1 * Math.PI / 180, // la, lo in radians
      la2 = lat2 * Math.PI / 180,
      lo1 = lon1 * Math.PI / 180,
      lo2 = lon2 * Math.PI / 180,
      haversine_phi = Math.pow(Math.sin((la2 - la1) / 2), 2),
      sin_lon = Math.sin((lo2 - lo1) / 2),
      h = haversine_phi + Math.cos(la1) * Math.cos(la2) * sin_lon * sin_lon;
    return 2 * R * Math.asin(Math.sqrt(h));
  }

  function toRad(angle) {
    return Math.PI * angle / 180;
  }

  function toDeg(angle) {
    return 180 * angle / Math.PI;
  }

  function constrain(input, min, max) {
    return (input > max) ? max : (input < min) ? min : input;
  }

  //azimuthal projection
  function toLocalCoordinates(lat, lon, ref_sin_lat, ref_cos_lat, ref_lon_rad) {
    var lat_rad = toRad(lat),
      lon_rad = toRad(lon),
      sin_lat = Math.sin(lat_rad),
      cos_lat = Math.cos(lat_rad),
      cos_d_lon = Math.cos(lon_rad - ref_lon_rad),
      c = Math.acos(constrain(
        ref_sin_lat * sin_lat + ref_cos_lat * cos_lat * cos_d_lon,
        -1,
        1
      )),
      k = (Math.abs(c) > 0) ? (c / Math.sin(c)) : 1,
      mul = k * R;

    return {
      "lat": mul * (ref_cos_lat * sin_lat - ref_sin_lat * cos_lat * cos_d_lon),
      "lon": mul * cos_lat * Math.sin(lon_rad - ref_lon_rad)
    };
  }

  function toGlobalCoordinates(lat, lon, ref_sin_lat, ref_cos_lat,
                               ref_lon_rad) {
    var x_rad = lat / R,
      y_rad = lon / R,
      c = Math.sqrt(x_rad * x_rad + y_rad * y_rad),
      sin_c = Math.sin(c),
      cos_c = Math.cos(c),
      lat_rad = Math.asin(
        cos_c * ref_sin_lat + (x_rad * sin_c * ref_cos_lat) / c
      ),
      lon_rad = ref_lon_rad + Math.atan2(
        y_rad * sin_c,
        c * ref_cos_lat * cos_c - x_rad * ref_sin_lat * sin_c
      );

    return {"lat": toDeg(lat_rad), "lon": toDeg(lon_rad)};
  }

  me.onStart = function () {
    var position = me.getCurrentPosition(),
      ref_lat_rad = toRad(position.x);

    me.direction_set = false;
    me.last_checkpoint_reached = false;
    me.next_checkpoint = 0;
    me.start_altitude = me.getInitialAltitude() + FLIGH_ALTITUDE;
    me.firstRun = true;

    me.ref_cos_lat = Math.cos(ref_lat_rad);
    me.ref_lon_rad = toRad(position.y);
    me.ref_sin_lat = Math.sin(ref_lat_rad);
  };

  me.onUpdate = function (timestamp) {
    var leaderYawRad;

    if (me.firstRun) {
      me.firstRun = false;
      me.secondRun = true;
      return;
    }
    if (me.secondRun) {
      me.leader_id = Math.trunc(me.drone_dict.length / 2);
      me.is_leader = me.id === me.leader_id;
      me.secondRun = false;
      me.neighbour_id = (me.id < me.leader_id) ? me.id + 1 : me.id - 1;
      me.angleUpdate = ((me.id < me.leader_id) ? 3 : 5) * Math.PI / 4;
    }
    if (!me.is_leader) {
      leaderYawRad = toRad(me.drone_dict[me.neighbour_id].yaw);

      me.neighbor_local_coordinates = toLocalCoordinates(
        me.drone_dict[me.neighbour_id].latitude,
        me.drone_dict[me.neighbour_id].longitude,
        me.ref_sin_lat,
        me.ref_cos_lat,
        me.ref_lon_rad
      );

      me.next_coordinates = toGlobalCoordinates(
        me.neighbor_local_coordinates.lat
          + MINIMAL_DISTANCE * Math.sin(leaderYawRad + me.angleUpdate),
        me.neighbor_local_coordinates.lon
          + MINIMAL_DISTANCE * Math.cos(leaderYawRad + me.angleUpdate),
        me.ref_sin_lat,
        me.ref_cos_lat,
        me.ref_lon_rad
      );

      return exitOnFail(
        me.setTargetCoordinates(
          me.next_coordinates.lat,
          me.next_coordinates.lon,
          me.drone_dict[me.neighbour_id].altitudeAbs
        ),
        "Failed to follow leader"
      );
    }

    if (!me.direction_set) {
      if (me.next_checkpoint < CHECKPOINT_LIST.length) {
        exitOnFail(
          me.setTargetCoordinates(
            CHECKPOINT_LIST[me.next_checkpoint].latitude,
            CHECKPOINT_LIST[me.next_checkpoint].longitude,
            CHECKPOINT_LIST[me.next_checkpoint].altitude + FLIGH_ALTITUDE
          ),
          "Failed to set checkpoint coordinates"
        );
        console.log("[DEMO] Going to Checkpoint " + me.next_checkpoint + "\n");
      }
      me.direction_set = true;
      return;
    }

    if (me.next_checkpoint < CHECKPOINT_LIST.length) {
      me.current_position = me.getCurrentPosition();
      me.distance = distance(
        me.current_position.x,
        me.current_position.y,
        CHECKPOINT_LIST[me.next_checkpoint].latitude,
        CHECKPOINT_LIST[me.next_checkpoint].longitude
      );
      if (me.distance > EPSILON) {
        console.log(
          "Waiting for drone to get to destination (checkpoint " +
            me.next_checkpoint + ": " + me.distance + " m)"
        );
      } else {
        console.log("[DEMO] Reached Checkpoint " + me.next_checkpoint + "\n");
        me.next_checkpoint += 1;
        me.direction_set = false;
      }
      return;
    }

    deployParachute(me);
  };

  me.onGetMsg = function (msg) {
    me.msgDict = JSON.parse(msg);

    if (msg === "landing") {
      deployParachute(me);
    } else {
      console.log("Unknown message: ", msg);
    }
  };
}(console, me));
